C  ====================================================================
C
C      Program  IDA   :  Inelastic Data Analysis
C      Modul    i72   :     tof kinematics
C
C  ====================================================================

C     Contents :
C
C        1.  Neutron kinematics:
C               yQ_of_0, yQ_of_W, Th2_of_W, E_of_l,
C               Tau_of_W, W_of_Tau, T_of_W, W_of_T, yKi_of_W
C
C        2.  Constant q interpolation:
C               CoQ

C  ====================================================================
C  i72 / 1 :   Neutron kinematics
C  ====================================================================

      REAL*8 FUNCTION yQ_of_0 ( E0, Th2)
C     ----------------------------------
         ! Calculate elastic wavenumber q for given scattering angle 2th
      REAL*8  K0, E0, Th2
      IF (E0.le.0) THEN
         Print *, 'yQ_of_0/ SEVERE ERROR/ E0<=0'
         yQ_of_0 = 0
         RETURN
         ENDIF
      K0 = dsqrt( E0 / 2.0723 )
      IF (Th2.lt.0) THEN
         Print *, 'yQ_of_0/ WARNING/ Th2<0'
         ENDIF
      yQ_of_0 = 2 * K0 * dsind ( Th2/2 )
      END ! yQ_of_0

      REAL*8 FUNCTION Th2_of_Q0 ( E0, Q)
C     ----------------------------------
         ! Calculate scattering angle 2th for given elastic wavenumber q
      REAL*8  K0, E0, Q, args
      K0 = dsqrt( E0 / 2.0723 )
      args = Q / 2 / K0
      IF (args.lt.-1 .or. args.gt.1) THEN 
         Print *, 'Th2_ofQ0/ WARNING/ q out of elastic range'
         Th2_of_Q0 = 0
      ELSE
         Th2_of_Q0 = 2 * dasind ( args )
         ENDIF 
      END ! Th2_of_Q0

      REAL*8 FUNCTION yQ_of_W ( W, E0, Th2 )
C     --------------------------------------
         ! Calculate inelastic q for scattering with 2th,w
      REAL*8  K, K0, W, E0, Th2
      IF (E0+W.lt.0.) THEN
         yQ_of_W = 0.
         Print *, 'invalid parameters in yQ_of_W : E0, W ', E0, W
         RETURN
         ENDIF
      K    = dsqrt( (E0 + W) / 2.0723 )
      K0   = dsqrt(  E0      / 2.0723 )
      yQ_of_W = dsqrt ( K**2 + K0**2 - 2*K*K0*dcosd(Th2) )
      END ! yQ_of_W

      REAL*8 FUNCTION Th2_of_W ( W, E0, Q )
C     -------------------------------------
         ! Calculate 2th for inelastic scattering with q,w
      REAL*8  Q, K, K0, W, E0, arg
      IF (E0+W.lt.0.) THEN
         Print *, ' Absturz in Th2_of_W : E0, W ', E0, W
         STOP
         ENDIF
      K    = dsqrt( (E0 + W) / 2.0723 )
      K0   = dsqrt(  E0      / 2.0723 )
      arg  = ( K**2 + K0**2 - Q**2 ) / (2*K*K0)
      IF (dabs(arg).gt.1.) THEN
         Print *, ' K K0 Q : ', K, K0, Q
         CALL Absturz ('Th2_of_W', '|arg|>+1')
      ELSE
         Th2_of_W = dacosd (arg)
         ENDIF
      END ! Th2_of_W

      REAL*8 FUNCTION E_of_l (lambda)
C     -------------------------------
         ! Convert neutron wavelength to energy
      REAL*8 lambda
      E_of_l = 81.805d0 / lambda**2
      END ! E_of_l

      REAL*8 FUNCTION Tau_of_W (W, E0)
C     --------------------------------
         ! time of flight for a neutron with energy E0+hquer*w
         ! since 21jun91 in milliseconds/meter, not microseconds !
      REAL*8 W, E0
      IF (W.le.-E0) THEN
         CALL Gong (17)
         Print *, ' BAD INPUT in Tau_of_W : w <= -E0'
         Tau_of_W = 1.
         RETURN
         ENDIF
      Tau_of_W = dsqrt ( 5.2271d0 / ( E0 + W ) )
      END ! Tau_of_W

      REAL*8 FUNCTION W_of_Tau (Tau, E0)
C     ----------------------------------
         ! energy gain E-E0 for a neutron with time of flight tau.
         ! msec/m -> meV
      REAL*8 Tau, E0
      IF (E0.eq.0.) THEN
         CALL Gong (17)
         CALL Absturz ('W_of_Tau', 'E0 = 0.0')
         ENDIF
      W_of_Tau =  5.2271d0 / Tau**2 - E0
      END ! W_of_Tau

      REAL*8 FUNCTION T_of_W (W)
C     --------------------------
         ! just t = hvoll/w [nsec / meV]
      REAL*8 W
      IF (W.le.0.) CALL Absturz ('T_of_W', 'W =< 0')
      T_of_W = 4.13570d-3 / W
      END ! T_of_W

      REAL*8 FUNCTION W_of_T (T)
C     --------------------------
         ! just w = hvoll/t [meV / nsec]
      IF (T.le.0.) CALL Absturz ('W_of_T', 'T =< 0')
      W_of_T = 4.13570d-3 / T
      END ! W_of_T

      REAL*8 FUNCTION DKi_of_W ( W, E0 )
C     ----------------------------------
         ! DKi = Ki(W) - Ki(W=0)
      REAL*8  K0, W, E0
      IF (E0+W.lt.0.) THEN
         Print *, ' Absturz in DKi_of_W : E0, W ', E0, W
         STOP
         ENDIF
      K0   = dsqrt(  E0      / 2.0723 )
      DKi_of_W = dsqrt (W/2.0723+ K0**2 )-K0
      END ! DKi_of_W

