// #include "myexception.h"
#include "workspace.h"
#include "file_io.h"

//! Load a .a01 formatted file.
void NFileIO::load_01( FILE *F_in, string flong )
{
    string lin, key, val, fdir, fshort, fext;
    uint iz, nj=0, j;
    CCoord co;
    SlicePtr sout;
    WorkFilePtr wout;

    mystd::fname_divide( flong, &fdir, &fshort, &fext);

    if     ( mystd::freadln(F_in, &lin)<=0 )
        throw BadData( "file empty or starting with empty line" );
    if     ( lin!=string("ASCII-01 JWu") )
        throw BadData( "a.01 file starts with illegal header [" + lin + "]" );

    if ( mystd::freadln(F_in, &lin)<=0 || lin.length()<5 || 
         lin.substr(0,4)!=string("typ ") )
        throw BadData( "no valid typ line" );
    if        ( lin.substr(4)=="D" ) {
        wout = WorkFilePtr( new DataFile() );
    } else if ( lin.substr(4)=="C" ) {
        wout = WorkFilePtr( new CurvFile() );
    } else {
        throw BadData( "! invalid type [" + lin.substr(4) + "]" );
    }

    while (1) { // loop over input lines

        if        ( mystd::freadln(F_in, &lin)<0 )
            throw BadData( "unexpected end-of-file" );

        // regular exit when "eof" is reached:
        if        (lin=="eof") {
            wout->as_on_disk = true;
            theWorkspace.push_back( wout );
            return;
        }

        if        ( lin.size() < 4 )
            throw BadData( "incomplete line" );

        key = lin.substr(0,4);
        val = lin.substr(4);

        if        (key=="fil ") {
            if (val!=fshort) {
                cout << "warning: file " << fshort <<
                    " renamed from " << val << "\n";
            }
            wout->name = fshort;
        } else if (key=="doc ") {
            wout->lDoc.push_back(val);
        } else if (key=="x   ") {
            wout->xco.load(val);
        } else if (key=="y   ") {
            wout->yco.load(val);
        } else if (lin[0]=='z') {
            if ( sscanf(lin.substr(1,3).c_str(), "%u", &iz)!=1 )
                throw BadData( "z coordinates must be numbered" );
            if ( iz!=wout->ZCo.size() )
                throw BadData( "z coordinates must be sorted" );
            co.load(val);
            wout->ZCo.push_back(co);
        } else if (key=="#j  ") {
            if ( sscanf(val.c_str(), "%u", &nj)!=1 )
               throw BadData( "after key #j no no. of scans\n");
        } else if (key=="scn ") {
            if ( sscanf(val.c_str(), "%u", &j)!=1 ) {
                cout << "warning: number of scan not given\n";
            } else if ( j!=wout->nK() ) {
                cout << "warning: scan " << wout->nK() <<
                    " has label " << j << "\n";
            }
            sout = wout->newSlice();
            loadSlice_01( F_in, sout, wout->nZ() );
            wout->V.push_back( sout );
        } else {
            throw BadData( "! Unexpected key [" + key + "]\n" );
        }
    }
}

//! Load a slices (aka "scan") from a .a01 formatted files.
void NFileIO::loadSlice_01( FILE *F_in, SlicePtr sout, int nz )
{
    int err;
    uint i, m, n, iv, iw;
    char lab[4];
    double v, w;

    // get z values:
    if ((err=fscanf(F_in, "%4c%u\n", lab, &m))!=2)
        throw BadData( "expecting Z head line, found " + strg(err) + 
                       " entries\n" );
    if (strncmp(lab, "Z   ", 4))
        throw BadData( "found [" + strg(lab) + "] in Z head line\n" );
    if (m!=nz)
        throw BadData( "found " + strg(m) + " Z entries, expecting " + 
                       strg(nz) + "\n" );
    for ( i=0; i<nz; ++i ) {
        if ((err=fscanf(F_in, "%lg\n", &v))!=1)
            throw BadData( "expecting z value, found " + strg(err) +
                           " entries\n" );
        sout->z.push_back( Value(v) );
    }
		
    if        (sout->styp()==DATA) {
        // get xy values:
        if ((err=fscanf(F_in, "%4c%u\n", lab, &n))!=2)
            throw BadData( "expecting XY head line, found " + strg(err) +
                           " entries\n" );
        if (strncmp(lab, "XY  ", 4))
            throw BadData( "found [" + strg(lab) + "] in XY head line\n" );
        Spectrum *spec = P_SPEC( sout );
        
        // Initialize with size 0, not n, because we are going to use
        // legacy_push_back. Presume that data are integer; this will
        // will bbe overwritten as soon as we call legacy_push_back_dble.
        spec->init_tuple( VAL_INT, 0 ); // maybe we have integer data ?
        
        char tx[21], ty[21];
        vector<string> txy(2);
        // cout << "going to read " << n << " entries\n";
        for (i=0; i<n; i++) {
            if ((err=fscanf(F_in, "%20s %20s\n", &tx, &ty))!=2)
                throw BadData( "did not find 20s 20s" );
            txy[0] = tx;
            txy[1] = ty;
            BOTH( ixy ){
                try{
                    iv = mystd::any2int( txy[ixy] );
                    // cout << i << ", " << ixy << " found integer " << iv << "\n";
                    spec->T[ixy]->legacy_push_back_int( iv );
                }
                catch(BadData &e){
                    v = mystd::any2dbl( txy[ixy] );
                    // cout << i << ", " << ixy << " found double " << v << "\n";
                    spec->T[ixy]->legacy_push_back_dbl( v );
                }
            }
        }
    } else if (sout->styp()==CURV) {
        throw NotImplemented( 
            "loadSlice_01: didnt know that curves were ever saved in a01" );
    } else {
        throw ProgramError( "unexpected styp in load_Slice_01" );
    }
}
