#pragma once

#include <vector>
#include <boost/shared_ptr.hpp>
#include <math.h>

#include "myexception.h"
#include "mystd.h"

/** Value and Channel are always of pure type (INT, DBL or ERR).
    Mixed type (MIX) and empty (NON) are only used when characterizing
    sets of Values or Channels. */
enum ValueType { VAL_INT, VAL_DBL, VAL_ERR, VAL_MIX, VAL_NON };

char valueTypeLetter( ValueType type );


//! Any (collection of) objects with well defined type (INT, DBL or ERR):
class Any {
 private:
    ValueType _vtyp;
 protected:
    void retype( ValueType type_in ) { _vtyp=type_in; };
 public:
    Any( ValueType type_in ) : _vtyp(type_in) {};
    ValueType vtyp() const { return _vtyp; };
    bool is_int() const { return _vtyp==VAL_INT; }
    bool is_dbl() const { return _vtyp==VAL_DBL || _vtyp==VAL_ERR; }
    bool has_err() const { return _vtyp==VAL_ERR; }
};


//! A Value is either int or double or double+variance.
/** All arithmetics in FRIDA is performed on values. */
class Value : public Any {
 private:
    int I;
    double D;
    double E; //!< despite the name e (error?), it's the variance (error^2)

 public:
    Value( int i_in ) : I(i_in), Any(VAL_INT) {};
    Value( double d_in ) : D(d_in), Any(VAL_DBL) {};
    Value( double d_in, double e_in ) : D(d_in), E(e_in), Any(VAL_ERR) {};

    int i() const { if( !is_int() ) throw ProgramError( "Value i()" ); return I; };
    double d() const { if( is_int() ) return (double)I; else if( !is_dbl() ) throw ProgramError( "Value d()" ); return D; };
    double e() const { return has_err() ? E : 0; };

    bool operator<  ( const Value &v2 ) const { return d()<v2.d(); };
    bool operator>  ( const Value &v2 ) const { return d()>v2.d(); };
    bool operator== ( const Value &v2 ) const { return is_int()&&v2.is_int() ? i()==v2.i() : d()==v2.d(); };
    bool operator<= ( const Value &v2 ) const { return !(*this>v2); };
    bool operator>= ( const Value &v2 ) const { return !(*this<v2); };
    bool operator!= ( const Value &v2 ) const { return !(*this==v2); };

    Value operator- () const;
    Value operator+ ( const Value &v2 ) const;
    Value operator- ( const Value &v2 ) const;
    Value operator* ( const Value &v2 ) const;
    Value operator/ ( const Value &v2 ) const;

    string strg() const {
        if( is_int() )
            return ::strg( i() );
        else if( has_err() )
            return ::strg( d() ) + "+-" + ::strg( sqrt(e()) );
        else
            return ::strg( d() );
    };
    //string str_d( const char* fmt );
    friend ostream& operator<< (ostream& s, Value v ) {
        if( v.is_int() )
            return s << v.i();
        else if( v.has_err() )
            return s << v.d() << "+-" << sqrt(v.e());
        else
            return s << v.d();
    };
};


//! A Channel is the optimized replacement for vector<Value>.
class Channel : public Any {
 public:
    vector<int>    II;
    vector<double> DD;
    vector<double> EE;
    Channel( ValueType type_in=VAL_DBL, uint n=0 ) : Any( type_in ) { resize(n); };

    void resize( uint n );
    void cast_to_double();
    void legacy_push_back_int( int iv );   //!< only for legacy read-in
    void legacy_push_back_dbl( double v ); //!< only for legacy read-in

    unsigned int size() const;
    double at (uint i) const { //!< slow data access
        if( is_int() ) return (double) II[i];
        else if( is_dbl() ) return DD[i];
        else throw ProgramError( "Channel [] invalid vtyp" );
    }
};

//! A ChannelPtr owns a Channel.
typedef boost::shared_ptr<Channel> ChannelPtr;
