#pragma once

// #include <iostream>
#include <vector>
#include <boost/shared_ptr.hpp>

#include "mystd.h"
#include "coord.h"
#include "value.h"
//#include "curve.h"
//#include "func.h"
//#include "expr.h"

// Down-casts SlicePtr to <Subclass>* (not elegant, but comparatively simple).
#define P_SPEC(pslice) (dynamic_cast<Spectrum*>(get_pointer(pslice)))
#define P_CURV(pslice) (dynamic_cast<Curve*>(get_pointer(pslice)))

// Down-casts WorkFilePtr to DataFile* or CurvFile*.
#define P_DFILE(pfile) (dynamic_cast<DataFile*>(get_pointer(pfile)))
#define P_CFILE(pfile) (dynamic_cast<CurvFile*>(get_pointer(pfile)))

// Loop x..y:
#define BOTH(ixy) for( uint ixy=0; ixy<2; ++ixy )



//! A slice is either a spectrum (DATA) or a curve (CURV).
/** Note however that it is good OO practice to avoid
    switch(type){..} whereever possible. Generally, it is
    better to call a virtual member function of Slice which
    is implemented differently for Spectrum and for Curve. */
enum SliceType { ANY_SLICE, DATA, CURV };


//! A Slice is either a Spectrum or a Curve (virtual baseclass).
class Slice {
 public:
    vector<Value> z;

    virtual SliceType styp() const =0;
};


//! An x-y data set, possibly with error bars
class Spectrum : public Slice {
 public:
    ChannelPtr T[2]; //!< Tuple (x-y)

    //Spectrum() : {};
    
    void init_channel( uint ixy, ValueType vtyp_in=VAL_DBL, uint n_in=0 ) { 
        T[ixy] = ChannelPtr( new Channel( vtyp_in, n_in ) ); };
    void init_tuple( ValueType vtyp_in=VAL_DBL, uint n_in=0 ) { 
        BOTH(ixy) init_channel( ixy, vtyp_in, n_in ); };

    void resize( uint n ){ BOTH(ixy) if( T[ixy] ) T[ixy]->resize( n ); }
    void clear(); 

    SliceType styp() const { return DATA; };
    void info() const;
    uint size() const;
};


//! A curve is a collection of parameters that allow to fit a spectrum.
class Curve : public Slice {
 public:
    vector<Value> quality;
    vector<Value> P;
    
    //    void eval( Spectrum *D, string &expr, CRange R, bool log, uint npts) const;
    SliceType styp() const { return CURV; };
    void info() const;
    void clear();
    double f(double x) const;
};


//! A SlicePtr owns a Slice.
typedef boost::shared_ptr<Slice> SlicePtr;


//! File hold in the workspace.
/** This is the virtual base class for DataFile and CurvFile. */
class WorkFile {
 public:
    // inline WorkFile() { cout << "DEBUG creating workfile\n"; };

    vector<SlicePtr> V;
    vector<CCoord> ZCo;
    vector<CParam> RPar;
    string name;
    vector<string> lDoc;
    CCoord xco, yco;
    uint id; //!< this is managed by class Workspace
    bool as_on_disk;

    uint nK() const { return V.size(); };
    uint nZ() const { return ZCo.size(); };

    bool has_data() const { return styp()==DATA; };
    bool has_curves() const { return styp()==CURV; };
    virtual SliceType styp() const =0;
    virtual SlicePtr newSlice() const =0;
};


//! Data file hold in the workspace.
class DataFile : public WorkFile {
 public:
    SliceType styp() const { return DATA; };
    SlicePtr newSlice() const { return SlicePtr( new Spectrum ); };
    ValueType channelType( uint ixy ) const;
    uint nT() const; //!< number of tuples per spectrum; (uint)(-1) unless unique
};


//! Curve file hold in the workspace.
class CurvFile : public WorkFile {
 public:
    string expr;
    //    CTree T;
    //uint no_dfil;
    vector<CCoord> PCo;
    vector<int>    Fixed; // should be vector<bool>, but trouble with IListV
    bool weight_y_log;

    SlicePtr newSlice() const { return SlicePtr( new Curve ); };
    SliceType styp() const { return CURV; };
    uint nPar() const { return PCo.size(); };
};


//! A pointer of this type _owns_ a WorkFile instance.
typedef boost::shared_ptr<WorkFile> WorkFilePtr;
	

typedef vector<WorkFilePtr> WorkFileVector;
// not sure whether we need to blow up this into a class definition:
//! Enhanced vector that knows about slice types.
// class WorkFileVector : public vector<WorkFilePtr> {
//  public:
//     WorkFileVector restrict( SliceType selected_type);
// }; 


//! This double-enhanced vector holds the complete workspace. 
class Workspace : public WorkFileVector {
 public:
    void push_back( WorkFilePtr pFile ) {
        WorkFileVector::push_back( pFile );
        pFile->id = size()-1;
    }
}; 

//! Singleton instance created in main module.
extern Workspace theWorkspace; 


//! Information about the files we are going to operate on:
class Job {
 public:
    WorkFileVector FileSet;
    WorkFileVector FileSetOld;
    WorkFileVector FileSetNew;
    bool force_inplace;
    bool force_copy;
};

//! Singleton instance created in main module.
extern Job theJob;


//! Loops over workfiles use this iterator.
class WorkFileIterator { 
 private:
    WorkFileVector *V;
    WorkFileVector::iterator p;
    SliceType our_styp;

    void restart() { p=V->begin(); };
    void check_styp();

 public:
    WorkFileIterator( SliceType selected_styp=ANY_SLICE,
                       WorkFileVector *V_in = &(theJob.FileSet) ) :
        our_styp(selected_styp), V (V_in) { restart(); check_styp(); };
    bool operator()( WorkFile **ppw );
    bool operator()( DataFile **ppwd );
};
