//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* axis.cpp: for plot                                                     *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <math.h>
#include <fcntl.h>
#include <boost/format.hpp>

#include <readln.h>

#include "mystd.h"
#include "axis.h"

using boost::format;


//! Set invalid limits.
//  This means: next plot call will automatically recalculate them.

void CAxis::setAuto()
{
    inf = -INFINITY;
    sup = +INFINITY;
}


//! Check and set limits.

void CAxis::setLimits( string axnam, double _inf, double _sup )
{
    if( _inf>=_sup )
        throw "Cannot set " + axnam + " plot range " +
            strg(_inf) + " .. " + strg(_sup);
    if( logflag && _inf<=0 )
        throw "Cannot set log " + axnam + " plot range " +
            strg(_inf) + " .. " + strg(_sup);
    inf = _inf;
    sup = _sup;
}


//! Set lin or log, reset ranges to auto.

void CAxis::setLog( bool _log )
{
    logflag = _log;
    setAuto();
}


//! Parse string to set bounds.

void CAxis::ask_and_set( const string& quest )
{
    for( ;; ) {
        double inf_in, sup_in;
        string resp1 = NRead::readwd( quest + " [" + str() + "] ? " );
        if        ( resp1 == "\\q" ) {
            throw "user escape to main menu";
        } else if ( resp1 == "" ) {
            return; // don't change current value
        } else if ( resp1 == "*" ) {
            setAuto();
            return;
        } else if ( resp1 == "\\EOL" ) {
            inf_in = inf;
        } else if( mystd::any2dbl(resp1,&inf_in) ) {
            ; // ok
        } else {
            cout << "required input: plot axis range\n"
                "examples:\n"
                "    0.1 1000    two floating-point numbers\n"
                "    \\   1000    let lower bond unchanged\n"
                "    0.1  \\      let upper bound unchanged\n"
                "    *           redetermine bounds from data\n"
                "just press ENTER to let present range unchanged [" 
                 << str() << "]\n";
            continue;
        }
        // cout << "DEBUG: inf_in = " << inf_in << "\n";
        if ( logflag && inf_in<=0 ) {
            cout << "log axis requires lower bound>0\n";
            continue;
        }
        string resp2 = NRead::readwd(
            ".. and upper limit [" + strg(sup) + "] ? " );
        if        ( resp2 == "\\q" ) {
            throw "user escape to main menu";
        } else if ( resp2 == "" || resp2 == "\\EOL" ) {
            sup_in = sup;
        } else if( mystd::any2dbl(resp2,&sup_in) ) {
            ; // ok
        } else {
            cout << "invalid upper bound; use '?' for help\n";
            continue;
        }
        if( sup_in<=inf_in ){
            cout << "invalid input, does not fulfill lower<upper\n";
            continue;
        }
        if( logflag && (sup_in>1e199 || inf_in<1e-199) ){
            cout << "log range currently restricted to 1E-199..1E199\n";
            continue;
        }
        if( !logflag && (sup_in>3e38 || inf_in<-3e38) ){
            cout << "lin range currently restricted to -3e38..3e38\n";
            continue;
        }
        inf = inf_in;
        sup = sup_in;
        return;
    }
}


//! Replace limits by rounded values.

void CAxis::setRoundedLimits( string axnam, double _inf, double _sup )
{
    static double relmargin = 0.05;
    static double digits = 2;

    if ( _inf==-INFINITY || _sup==+INFINITY )
        throw "BUG: rounding " + axnam + " limits: unexpected inf limits";
    if( _inf>_sup )
        throw "BUG: rounding " + axnam + " limits: invalid plot range " +
            strg(_inf) + " > " + strg(_sup);
    if( logflag && _inf<=0 )
        throw "BUG: rounding " + axnam + " limits: nonpositive log inf " +
            strg(_inf);

    inf = _inf;
    sup = _sup;
                      
    if ( inf==sup ) {
        if( logflag ){
            inf /= 10;
            sup *= 10;
        } else if ( inf>0 ) {
            inf *= 0.9;
            sup *= 1.1;
        } else if ( sup<0 ) {
            inf *= 1.1;
            sup *= 0.9;
        } else {
            inf = -1;
            sup = +1;
        }
        return;
    } 

    double inf2, sup2;
    double mydigits = digits;
    if ( !logflag ) { // lin limits
        double margin = relmargin * (sup - inf);
        do {
            if (mydigits>=8.5) return; // do not round
            sup2 = mystd::round_decimal( sup+margin, mydigits );
            inf2 = mystd::round_decimal( inf-margin, mydigits );
            if(sup<0 && sup2>0) sup2 = 0;
            if(inf2<0 && inf>0) inf2 = 0;
            mydigits += 0.5;
        } while (!((inf2<inf) && (sup<sup2)));
    } else { // log limits
        double ratio = sup / inf;
        double margin = exp( relmargin*log(ratio) );
        do {
            if (mydigits>=8.5) return; // do not round
            sup2 = mystd::round_decimal( sup*margin, mydigits );
            inf2 = mystd::round_decimal( inf/margin, mydigits );
            mydigits += 0.5;
        } while ( !((inf2<inf) && (sup<sup2)) );
    }
    setLimits( axnam, inf2, sup2 );
}	


//! Have finite limits been set ?

bool CAxis::finite() const
{
    return inf!=-INFINITY && sup!=+INFINITY;
}


//! Is value val contained in this range?

bool CAxis::contains(double val)  const
{
    return inf<=val && val<=sup;
}


//! Map application scale (inf..sup) to linear plot scale (0..1).

double CAxis::value2plotcoord( double v ) const
{
    if ( !finite() )
        throw string( "undefined plot range" );
    if ( logflag ) {
        if( inf<0 || v<0 )
            throw string( "negative value in log range" );
        return log(v/inf) / log(sup/inf);
    } else {
        return (v-inf) / (sup-inf);
    }
}


//! Map linear plot scale (0..1) to application scale (inf..sup).

double CAxis::plotcoord2value( double c ) const
{
    if ( !finite() )
        throw string( "undefined plot range" );
    if ( logflag ) {
        return inf * exp( c*log(sup/inf) );
    } else {
        return inf + c*(sup-inf);
    }
}


//! String representation.

string CAxis::str() const
{
    if( inf==-INFINITY || sup== +INFINITY )
        return "*";
    else
        return strg(inf) + " " + strg(sup);
}


//! Describe range, for use in table of plot frames.

string CAxis::info() const
{
    string ret;
    ret  =   logflag ? "log" : "lin";
    ret += " " + strg(inf);
    ret += " " + strg(sup);
    return ret;
}


//! Convert value -> plot_coordinate (my PostScript units 0..10).

double CAxis::pc( double v ) const
{
    return 10 * value2plotcoord( v );
}


//! Calculate appropriate division of plot axis.

void CAxis::calc_ticks( int *ntack, double **tack, int *ntpt, double *ticklim)
    const
{
    int ir, nt, i;
    double r, rd, rdr, d, vsub, vsup, tack0, tackn, dtack;
	
    if ( inf >= sup )
        throw "PROGRAM ERROR/ Ticks/ Bad Limits: " +
            strg(inf) + ", " + strg(sup);

    if (!logflag) {
        r = log10(sup - inf);
        ir = (int) ((r<0) ? r-1 : r);
        rd = r - ir; // fractional part of r
        rdr = pow(10., rd);
        if        (rdr > 9.99999) {
            dtack = 2.5 * pow(10., ir); // spacing between tacks
            *ntpt = 5;                // ticks per tack
        } else if (rdr > 5.00001) {
            dtack = 2 * pow(10., ir);
            *ntpt = 4;
        } else if (rdr > 2.500005) {
            dtack = 1 * pow(10., ir);
            *ntpt = 5;
        } else if (rdr > 1.600003) {
            dtack = .5 * pow(10., ir);
            *ntpt = 5;
        } else if (rdr > 1.250002) {
            dtack = .4 * pow(10., ir);
            *ntpt = 4;
        } else {
            dtack = .25 * pow(10., ir);
            *ntpt = 5;
        }
        d = inf / dtack;
        tack0 = dtack * (int) (d<=0?d-1:d);
        d = sup / dtack;
        tackn = dtack * ((int) (d<0?d-1:d) + 1);
        vsub = inf - (sup-inf)*1.e-4;
        vsup = sup + (sup-inf)*1.e-4;
        nt = (int) ((tackn-tack0) / dtack) + 3; // allocate enough
        *tack = (double*) malloc(sizeof(double)*nt);
        *ntack = 0;
        for (i=0; i<nt; i++) {
            d = tack0 + i*dtack;
            if (vsub<=d && d<=vsup)
                (*tack)[(*ntack)++] = d;
        }

        ticklim[0] = tack0;
        ticklim[1] = (*tack)[(*ntack)-1] + dtack;
		
    } else { // log scale

        static double eins = 1.0000009999999999;
        double rlgmin, rlgmax, rlgrel, d1, d2, r0; 
        int    minexp, maxexp, incr;

        if (inf <= 0)
            throw string( "PROGRAM ERROR/ Ticks/ negative log argument" );

        rlgmin = log10(inf);
        rlgmax = log10(sup);
        rlgrel = rlgmax - rlgmin;
        if (rlgrel > 40)
            throw string( "Excessive log range" );

        minexp = (rlgmin>0 ? 
                  (int) (rlgmin+1e-6) :
                  ((int) (rlgmin-1e-6) -1));
        maxexp = (rlgmax>0 ? 
                  (int) (rlgmax+1e-6) :
                  ((int) (rlgmax-1e-6) -1));

        incr = (int) (rlgrel / 7 + 1); // #decades per tack
        if (incr==2) incr = 1; // preferences: 10^1, 10^3, 10^6 etc.
        if (incr>3) incr = 3 * (incr % 3);

        if (incr > 1) {
            minexp -= minexp % incr;
            maxexp -= maxexp % incr;
        }

        /*  - Set large ticks array : */
        static int mt=30;
        *tack = (double*) malloc(sizeof(double)*mt);

        *ntack = 0;
        d1 = inf / eins;
        d2 = sup * eins;
        for ( i = minexp; incr < 0 ? i >= maxexp : i <= maxexp; i += incr ) {
            r0 = pow(10., i);
            if (d1<=r0 && r0<=d2) {
                if(*ntack>=mt)
                    throw string( "PROG ERR too many tacks" );
                (*tack)[(*ntack)++] = r0;
            }
        }

        if (*ntack >= 1) {
            ticklim[0] = (*tack)[0]          / pow(10., incr);
            ticklim[1] = (*tack)[(*ntack)-1] * pow(10., incr);
        } else {
            // untested. just to allow plots when no tacks in window
            ticklim[0] = pow( 10., minexp );
            ticklim[1] = pow( 10., maxexp+1 );
        }

        /** determine # ticks / tack **/
        if        (incr==1 && rlgrel<=7) {
            *ntpt = 9;
        } else if (incr==1 && rlgrel<=7) {
            *ntpt = 3; // 1-2-5-10 - Schritte 
        } else if (incr==1) {
            *ntpt = 1;
        } else {
            *ntpt = -incr; 
            // means incr ticks per tack
            // means 1 tick per decade
        }
    }
}
