#include <boost/shared_ptr.hpp>

#include "list.h"
#include "coord.h"
#include "zentry.h"


//! Shared pointers, used in function declarations.

typedef boost::shared_ptr<class COlo> POlo;
typedef boost::shared_ptr<class COld> POld;
typedef boost::shared_ptr<class COlc> POlc;

typedef boost::shared_ptr<class CTree> PTree;


//! Online object: virtual base class for COld, COlc.

class COlo {
 public:
    // data records:
    vector<PZentry> V;
    string name;
    vector<string> lDoc;
    CCoord xco, yco;
    vector<CCoord> ZCo;
    vector<CParam> RPar;

    // online state records:
    bool as_on_disk;
    int test;
    CList ProtectedSpecs;

    COlo() : as_on_disk(false) {};
    virtual ~COlo() {}; // magic: makes dynamic_cast possible

    void clear() { V.clear(); };
    void remove( uint j ) { V.erase( V.begin()+j ); };
    void remove_z( uint iz ) {
        ZCo.erase( ZCo.begin()+iz );
        for( uint j=0; j<nJ(); ++j )
            V[j]->z.erase( V[j]->z.begin()+iz );
    };
    virtual void copy_mainvec( POlo fin ) = 0;

    uint nJ() const { return V.size(); };
    uint nZ() const { return ZCo.size(); };
    double z( uint j, uint iz ) const { return V[j]->z[iz]; };
    virtual PZentry new_zentry( uint j ) const = 0;
    virtual POlo new_olo() const = 0;
};


//! Online data file.

class COld : public COlo {
 public:
    COld() {};
    COld( class COlc const* fc );

    // trivially duplicated functions:
    PSpec VS( uint j ) const;
//    void remove_z( uint iz ){ COlo::remove_z( iz ); };
    PZentry new_zentry( uint j ) const;
    POlo new_olo() const;
    POld new_old() const;
    void copy_mainvec( POlo fin );

    // pertinent functions:
    uint nPts(uint j) const; 
    uint nPts() const; // 0 unless all spectra j have same nPts(j)
};


//! Online curve file.

class COlc : public COlo {
 public:
    // basic records (offline and online):
    string expr;
    vector<CCoord> PCo;

    // online state records:
    PTree T;
    enum TWgt { _LIN, _LOG, _ERR } weighing;
    uint           kd; // internal number of data file to be fitted
    uint           kconv;
    uint           cv_intmod;
    double         cv_epsabs;
    double         cv_epsrel;

    string range_expr;
    PTree range_T;

    COlc() : weighing( _ERR ), kconv(-1), range_expr("") { ; };

    // trivially duplicated functions:
    PCurve VC( uint j ) const;
    PZentry new_zentry( uint j ) const;
    POlo new_olo() const;
    POlc new_olc() const;
    void copy_mainvec( POlo fin );

    // pertinent functions:
    uint nPar() const { return PCo.size(); };
    string weight_str() const;
    vector<string> pInfo() const;
    string pInfoCat() const;
};


//! Casts.

#define P2D(p) boost::dynamic_pointer_cast<COld>(p)
#define P2C(p) boost::dynamic_pointer_cast<COlc>(p)
